from __future__ import print_function, unicode_literals, absolute_import
from twisted.web import http


class RtplotHandler(http.Request):
    """
    Server for requests from rtplot.
    The response delivers the binning factors, number of windows and
    their positions.
    """
    def process(self):
        "Send window params."
        try:
            self.setHeader('Content-Type', 'text/plain')
            wins = self.channel.instpars.getRtplotWins()
            if wins == '':
                self.write('No valid data available\r\n'.encode())
            else:
                self.write(wins.encode())
            self.finish()
        except Exception as err:
            self.channel.globals.clog.warn('RtplotServer: ', err)


class RtplotChannel(http.HTTPChannel):
    requestFactory = RtplotHandler

    def __init__(self, instpars, globals):
        http.HTTPChannel.__init__(self)
        self.instpars = instpars
        self.globals = globals


class RtplotFactory(http.HTTPFactory):
    def __init__(self, instpars, globals, **kwargs):
        self.instpars = instpars
        self.globals = globals
        http.HTTPFactory.__init__(self, **kwargs)

    def buildProtocol(self, addr):
        return RtplotChannel(self.instpars, self.globals)



# read in config
from __future__ import absolute_import, print_function, division
import configobj
import os
import validate

from twisted.internet.defer import inlineCallbacks

from hcam_widgets.misc import createJSON, saveJSON

try:
    from importlib import resources as importlib_resources
except Exception:
    # backport for python 3.6
    import importlib_resources


def check_user_dir(g):
    """
    Check directories exist for saving apps/configs etc. Create if not.
    """
    dirs = [os.path.expanduser("~/.hdriver")]
    if "app_directory" in g.cpars:
        dirs.append(os.path.expanduser(g.cpars["app_directory"]))
    if "log_file_directory" in g.cpars:
        dirs.append(os.path.expanduser(g.cpars["log_file_directory"]))
    for direc in dirs:
        if not os.path.exists(direc):
            try:
                os.mkdir(direc)
            except Exception as err:
                g.clog.warn("Failed to make directory " + str(err))


def load_config(g):
    """
    Populate application level globals from config file
    """
    configspec_file = str(
        importlib_resources.files("hcam_drivers") / "data" / "configspec.ini"
    )
    # try and load config file.
    # look in the following locations in order
    # - ~/.hdriver directory
    # - package resources
    paths = []
    paths.append(os.path.expanduser("~/.hdriver/"))
    resource_dir = str(importlib_resources.files("hcam_drivers") / "data")
    paths.append(resource_dir)

    # now load config file
    config = configobj.ConfigObj({}, configspec=configspec_file)
    for loc in paths:
        try:
            with open(os.path.join(loc, "config")) as source:
                config = configobj.ConfigObj(source, configspec=configspec_file)
            break
        except IOError:
            pass

    # validate ConfigObj, filling defaults from configspec if missing from config file
    validator = validate.Validator()
    result = config.validate(validator)
    if result is not True:
        g.clog.warn("Config file validation failed")

    # now update globals with config
    g.cpars.update(config)


def write_config(g):
    """
    Dump application level globals to config file
    """
    configspec_file = str(
        importlib_resources.files("hcam_drivers") / "data" / "configspec.ini"
    )
    config = configobj.ConfigObj({}, configspec=configspec_file)
    config.update(g.cpars)
    config.filename = os.path.expanduser("~/.hdriver/config")
    try:
        config.write()
    except Exception as err:
        g.clog.warn("Could not write config file:\n" + str(err))


@inlineCallbacks
def dump_app(g):
    """
    Dump current application settings to backup
    """
    json_string = yield createJSON(g, full=False)
    saveJSON(g, json_string, backup=True)



# Some utility tools to handle rising and setting etc.
from __future__ import print_function, unicode_literals, absolute_import, division
import traceback
import struct
import os
import json

import numpy as np
import yaml
from tornado.web import RequestHandler
from tornado.escape import json_encode
from astropy.io import fits
from astropy.utils.decorators import lazyproperty
from six.moves import urllib


MSG_TEMPLATE = "MESSAGEBUFFER: {}\nRETCODE: {}"
FRAME_NUMBER_URL = 'http://localhost:5000/status/DET.FRAM2.NO'


def getLastFrameNumber():
    """
    Polls the Hipercam server to find the current frame number

    Raises an exception in cases of failure
    """
    response = urllib.request.urlopen(FRAME_NUMBER_URL, timeout=0.5).read()
    try:
        data = json.loads(response)
    except Exception:
        raise Exception('cannot parse server response')
    if data['RETCODE'] != 'OK':
        raise Exception('server response not OK')
    msg = data['MESSAGEBUFFER']
    try:
        frame_number = int(msg.split()[1])
    except Exception:
        raise Exception('getLastFrameNumber error: msg = ' + msg)
    return frame_number


class BaseHandler(RequestHandler):
    """
    Abstract class for request handling
    """
    def initialize(self, db):
        self.db = db

    def write_error(self, status_code, **kwargs):
        self.set_header('Content-Type', 'application/json')
        resp = MSG_TEMPLATE.format(self._reason, 'NOK')
        resp_dict = yaml.load(resp)
        if self.settings.get("serve_traceback") and "exc_info" in kwargs:
            lines = []
            for line in traceback.format_exception(*kwargs["exc_info"]):
                lines.append(line)
            resp_dict['traceback'] = lines
        self.finish(json_encode(resp_dict))


class FastFITSPipe:
    def __init__(self, fileobj):
        """
        Simple class to quickly read raw frame bytes from HiPERCAM FITS cube.

        For example::

            >> fileobj = open('example.fits', 'rb')
            >> ffp = FastFITSPipe(fileobj)
            >> ffp.seek_frame(100)
            >> hdu_data = ffp.read_frame_bytes()

        Parameters
        -----------
        fileobj : file-like object or str
            the file-like object representing a FITS file, readonly
        """
        # assume fileobj is string
        try:
            self._fileobj = open(fileobj, 'rb')
        except Exception:
            self._fileobj = fileobj
        self._header_bytesize = None
        self.dtype = np.dtype('int16')

    @property
    def num_frames(self):
        # first, see if it's in the headers
        try:
            num = self.hdr['NAXIS3']
            if num == 0:
                raise ValueError('run still in progress')
        except (KeyError, ValueError) as err:
            # try and use ESO fileserver if it is running
            try:
                num = getLastFrameNumber()
            except Exception:
                # last, desperate, chance to try to guess from the filesize
                current_size = os.stat(self._fileobj.name).st_size
                # cant use integer division because timestamps are buffered and 2800 fits
                # block size causes trouble. Also, for some crazy reason the filesize
                # from os.stat seems to be the header size and multiple of the frame
                # size *without the timing bytes*. I don't understand this, but still...
                data_size = current_size - self.header_bytesize
                num = int(round(data_size / (self.framesize - 36)))
        return num

    @lazyproperty
    def hdr(self):
        self._fileobj.seek(0)
        return fits.Header.fromfile(self._fileobj)

    @property
    def header_bytesize(self):
        if self._header_bytesize is None:
            self._fileobj.seek(0)
            _ = fits.Header.fromfile(self._fileobj)
            self._header_bytesize = self._fileobj.tell()
        return self._header_bytesize

    @lazyproperty
    def framesize(self):
        # get nsamples per pixel, old format should default to 1
        # since NX and NY are always right in old format, but
        # NX is four times too high when NSAMP = 4
        nsamp = self.hdr.get('ESO DET NSAMP', 1)
        size = 18 + (self.hdr['ESO DET ACQ1 WIN NX'] * self.hdr['ESO DET ACQ1 WIN NY']) // nsamp
        bitpix = self.hdr['BITPIX']
        size = abs(bitpix) * size // 8
        # currently metadata consists of 36 bytes per frame (for timestamp)
        return size

    def seek_frame(self, frame_number):
        """
        Try and find the start of a given frame

        Raises exception if frame not written yet
        """
        self._fileobj.seek(self.header_bytesize + self.framesize*(frame_number-1))

    def read_frame_bytes(self):
        start_pos = self._fileobj.tell()
        raw_bytes = self._fileobj.read(self.framesize)
        if len(raw_bytes) != self.framesize:
            # go back to start position
            self._fileobj.seek(start_pos)
            raise EOFError('frame not written yet')
        return raw_bytes


def raw_bytes_to_numpy(raw_data, bscale=1, bzero=32768, dtype='int16'):
    """
    Convert output from FastFITSPipe to numpy array

    For example::

        >> raw_data = ffp.read_frame_bytes()
        >> data = raw_bytes_to_numpy(raw_data)
        >> data = data.reshape(ffp.output_shape)

    Parameters
    -----------
    raw_data : bytes
        bytes returned from FastFITSPipe
    bscale : int, default = 1
        scaling to apply to raw data.
        FITS cannot stored unsigned 16-bit integers, so
        data is usually stored as signed 16-bit and then scaled
    bzero : int, default = 32768
        offset to apply to raw data
    dtype : string, default='int16'
        data type used to store data
    """
    data = np.fromstring(raw_data, np.dtype(dtype))
    data.dtype = data.dtype.newbyteorder('>')
    np.multiply(data, bscale, data)
    # now try and recast back to unsigned ints without a copy
    # see stackoverflow.com/questions/4389517/in-place-type-conversion-of-a-numpy-array
    view = data.view('uint16')
    view.dtype = view.dtype.newbyteorder('>')
    view += bzero
    return view


def decode_timestamp(ts_bytes):
    """
    Decode timestamp tuple from values saved in FITS file

    The timestamp is originally encoded to bytes as a series of
    32bit (4 bytes) unsigned integers in little endian byte format.

    Added to this is a pair of chars (1 byte) for the number of satellites
    tracked and a sync status.

    However, this is then stored as fake pixels in a FITS file, which
    performs some mangling of the data, since FITS assumes 16bit (2 byte)
    integers, and has no way to store unsigned integers. The following
    mangling occurs:

    - decode the timestamp byte string as two 16bit (2 bytes) little-endian unsigned integers;
    - subtract 32768 from each integer;
    - encode these integers as two 16bit signed integers in BIG ENDIAN format;
    - save to file as fits data.

    This routine reverses this process to recover the original timestamp tuple. We have to
    take some care because of all the endian-swapping going on. For example, the number 27
    starts off as \x1b\x00\x00\x00, which is interpreted by the FITS save process as (27, 0).
    If we ignore the signing issue for clarity, then (27, 0) encoded in big endian format is
    \x00\x1b, \x00\x00 so we've swapped the byte pairs around.

    The steps taken by this routine are:

    - decode timestamp string as big endian 16bit integers
    - add 32768 to each value
    - encode these values as little endian 16bit unsigned integers
    - re-interpret the new byte string as 32bit, little-endian unsigned integers

    The status bytes are handled slightly differently: after the re-encoding to little endian
    16 bit uints, they are decoded as chars, and the last two chars discarded.

    Parameters
    ----------
    ts_bytes: bytes
        a Python bytes object which contains the timestamp bytes as written in the
        FITS file

    Returns
    --------
    timestamp : tuple
        a tuple containing the (frameCount, timeStampCount,
                                years, day_of_year, hours, mins,
                                seconds, nanoseconds) values.
    """
    buf = struct.pack('<' + 'H'*18, *(val + 32768 for val in struct.unpack('>'+'h'*18, ts_bytes)))
    return struct.unpack('<' + 'I'*8, buf[:-4]) + struct.unpack('bb', buf[-4:-2])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup
import glob
import os
import sys

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = [
    "astropy",
    "autobahn",
    "tornado",
    "pyaml",
    "twisted",
    "configobj",
    "tqdm",
    "hcam_widgets>=1.1.2",
    "hcam_devices>=1.1.1",
]
if sys.version_info[0] == 2:
    requirements.append("pymodbus")
else:
    requirements.append("pymodbus3")

test_requirements = [
    # TODO: put package test requirements here
]

# Treat everything in scripts except README.rst as a script to be installed
scripts = [
    fname
    for fname in glob.glob(os.path.join("scripts", "*"))
    if os.path.basename(fname) != "README.rst"
]

setup(
    name="hcam_drivers",
    version="1.4.1",
    description="Observation planning and finding charts for HiPerCAM",
    long_description=readme + "\n\n" + history,
    author="Stuart Littlefair",
    author_email="s.littlefair@shef.ac.uk",
    url="https://github.com/HiPERCAM/hcam-drivers",
    download_url="https://github.com/HiPERCAM/hcam-drivers/archive/v1.4.1.tar.gz",
    packages=["hcam_drivers", "hcam_drivers.utils"],
    package_dir={"hcam_drivers": "hcam_drivers"},
    include_package_data=True,
    scripts=scripts,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords="hcam-drivers",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
    ],
    test_suite="tests",
    tests_require=test_requirements,
)



# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
from __future__ import print_function, absolute_import, unicode_literals, division

import numpy as np
from astropy.coordinates.matrix_utilities import rotation_matrix


def calculate_sky_offset(xoff, yoff, sky_pa):
    """
    Convert pixel offset to sky offset in arcseconds

    Arguments
    ---------
    xoff, yoff : float
        desired pixel offsets
    sky_pa : float
        current instrument PA
    """
    px_scale = 0.081  # arcseconds per pixel
    flipEW = True  # is E to the right in the image?
    EofN = True  # does increasing rotator PA move us to E?
    paOff = 209.7 # pa when rotator = 0

    if EofN:
        theta = sky_pa - paOff
    else:
        theta = -sky_pa - paOff
    # only take 3d of 3d rotation matrix
    rmat = rotation_matrix(theta)[:2, :2]

    # +ve shifts should move stars right and up
    # offset for skyPA = 0 are:
    ra_shift_arcsecs = -xoff*px_scale if flipEW else xoff*px_scale
    dec_shift_arcsecs = -yoff*px_scale
    pix_shift = np.array([ra_shift_arcsecs, dec_shift_arcsecs])
    return rmat.dot(pix_shift)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

# This sub-module is destined for common non-package specific utility
# functions that will ultimately be merged into `astropy.utils`

# As far as possible, utils contains classes of generic use, such as
# PosInt for positive integer input. See e.g. 'hcam' for more instrument
# dependent components.

from __future__ import print_function, unicode_literals, absolute_import, division

__author__ = """Stuart Littlefair"""
__email__ = 's.littlefair@shef.ac.uk'
